module Emitter where

import List

import Parser
import Table
import Instruction

data Environment = Environment (Table, Int, Int) deriving Show

envTable (Environment (table, offset, label)) = table
envOffset (Environment (table, offset, label)) = offset
envLabel (Environment (table, offset, label)) = label
incEnvOffset (Environment (table, offset, label)) = 
    if offset < 0 then
        emitError "memory over"
    else
        (Environment (table, offset - 1, label))
decEnvOffset (Environment (table, offset, label)) = 
    (Environment (table, offset + 1, label))
incEnvLabel (Environment (table, offset, label)) = (Environment (table, offset , label + 1))

data EmitValue =
    EmitVoidValue |
    EmitIntValueTemporary Int |
    EmitIntVariable Int String |
    EmitIntValueConstant Int |
    EmitValueFalseLabels [String] |
    EmitValueTrueLabels [String] |
    EmitIntValueRegisterA |
    EmitAddressRegisterY |
    EmitNone deriving (Show)

data EmitResult =
    EmitResult (EmitValue, [(Instruction, String)], Environment) |
    EmitResultAlternative EmitResult [EmitResult]
    deriving Show

emitResultValue (EmitResult (ev, code, env)) = ev
emitResultValue (EmitResultAlternative er _) = emitResultValue er
emitResultCode (EmitResult (ev, code, env)) = code
emitResultCode (EmitResultAlternative er _) = emitResultCode er
emitResultEnvironment (EmitResult (ev, code, env)) = env
emitResultEnvironment (EmitResultAlternative er _) = emitResultEnvironment er

emitResultSelect p er@(EmitResult _) = er
emitResultSelect p (EmitResultAlternative der (er:ers)) | p er = er
emitResultSelect p (EmitResultAlternative der (er:ers)) = emitResultSelect p (EmitResultAlternative der ers)
emitResultSelect p (EmitResultAlternative der []) = der

emitResultIsFalseLabels (EmitResult (EmitValueFalseLabels _, _, _)) = True
emitResultIsFalseLabels _ = False

emitResultIsTrueLabels (EmitResult (EmitValueTrueLabels _, _, _)) = True
emitResultIsTrueLabels _ = False


emitError = error . ("emitter error: " ++)

emit tree = emitProgram tree (Environment (initialTable "undeclared variable ", 0xf, 0))

emitProgram (Program functions) env = emitConcat emitFunction env functions
emitConcat f env l = concat $ map (\x -> f x env) l

emitFunction (Function (name, ftype, arguments, block)) env =
    emitResultCode (emitBlock block env)

emitBlock (Block (declarations, statements)) env =
    EmitResult(emitResultValue statementsResult,
               emitResultCode statementsResult,
               Environment (envTable env, envOffset env, envLabel (emitResultEnvironment statementsResult)))
    where
      newEnv = emitDeclarations declarations env
      statementsResult = emitStatements statements newEnv

emitStatements = emitSequence emitStatement

emitSequence ef (t:ts) env = 
    EmitResult (if ts == [] then EmitNone else emitResultValue emitResult,
                (emitResultCode emitResult) ++ emitResultCode nextEmitResult,
                emitResultEnvironment nextEmitResult)
    where
      emitResult = ef t env
      nextEmitResult = emitSequence ef ts (emitResultEnvironment emitResult)
emitSequence ef [] env = EmitResult (EmitNone, [], env)

emitStatement tree env = 
    case tree of
      (Substitution ((Variable variable), expression)) ->
          let 
              emitResult = emitExpression expression env
              emitVal = emitResultValue emitResult
              code = emitResultCode emitResult
              newEnv = emitResultEnvironment emitResult
              variableOffset = offsetTable (envTable env) variable
          in
            EmitResult(EmitNone,
                       code ++ 
                       toRegisterA emitVal ++
                       [(OperandInt1 "TIY" variableOffset, variable ++ " = ..."),
                        (Operand0 "AM", "")],
                       newEnv)
      (If (expression, statements)) -> emitStatementIf (emitExpression expression env) statements

      (IfElse (expression, thenStatements, elseStatements)) -> emitStatementIfElse (emitExpression expression env) thenStatements elseStatements

      (While (expression, statements)) -> emitStatementWhile (emitExpression expression env) statements

      (InfiniteLoop (statements)) ->
          let
              emitResult = emitStatement statements env
              newEnv = emitResultEnvironment emitResult
              labelNo = envLabel newEnv
              label1 = emitLabel labelNo
          in
            EmitResult(EmitNone,
                       [(Label label1, "infinite loop start")] ++
                       emitResultCode emitResult ++
                       [(FlagResetDontCare, ""),
                        (OperandString1 "JUMP" label1, "infinite loop jump")],
                       incEnvLabel newEnv)

      (StatementExpression expression) -> emitExpression expression env

      (StatementBlock block) -> emitBlock block env

      (StatementEmpty) -> EmitResult (EmitNone, [], env)

      _ -> emitError ("unkown statement")

emitStatementIf (EmitResult (condVal, condCode, condEnv)) statements =
    let
        thenResult = emitStatement statements condEnv
        newEnv = emitResultEnvironment thenResult
        labelNo = envLabel newEnv
        label1 = emitLabel labelNo
    in
      EmitResult
      (EmitNone,
       condCode ++ toRegisterA condVal ++
       [(OperandInt1 "AIA" 0xf, "if judege"),
        (OperandInt1 "AIA" 1, ""),
        (OperandString1 "JUMP" label1, "")] ++ 
       emitResultCode thenResult ++
       [(Label label1, "if end")],
       incEnvLabel newEnv)
emitStatementIf (EmitResultAlternative der (EmitResult (EmitValueFalseLabels condLabels, condCode, condEnv):ers)) statements =
    let
        thenResult = emitStatement statements condEnv
        newEnv = emitResultEnvironment thenResult
    in                      
      EmitResult
      (EmitNone,
       condCode ++
       emitResultCode thenResult ++
       emitLabels condLabels "if end",
       newEnv)
emitStatementIf (EmitResultAlternative der []) statements = emitStatementIf der statements
emitStatementIf (EmitResultAlternative der (er:ers)) statements = emitStatementIf (EmitResultAlternative der ers) statements

emitStatementIfElse (EmitResult (condVal, condCode, condEnv)) thenStatements elseStatements =
    let
        thenResult = emitStatement thenStatements condEnv
        elseResult = emitStatement elseStatements (emitResultEnvironment thenResult)
        newEnv = emitResultEnvironment elseResult
        labelNo = envLabel newEnv
        label1 = emitLabel labelNo
        label2 = emitLabel (labelNo + 1)
    in
      EmitResult
      (EmitNone,
       condCode ++ toRegisterA condVal ++
       [(OperandInt1 "CIA" 0, "if else judge"),
        (OperandString1 "JUMP" label1, "")] ++ 
       emitResultCode elseResult ++
       [(FlagResetDontCare, ""),
        (OperandString1 "JUMP" label2, "if else jump"),
        (Label label1, "")] ++
       emitResultCode thenResult ++
       [(Label label2, "if else end")],
       (incEnvLabel . incEnvLabel) newEnv)
emitStatementIfElse (EmitResultAlternative der (EmitResult (EmitValueFalseLabels condLabels, condCode, condEnv):ers)) thenStatements elseStatements =
    let
        thenResult = emitStatement thenStatements condEnv
        elseResult = emitStatement elseStatements (emitResultEnvironment thenResult)
        newEnv = emitResultEnvironment elseResult
        labelNo = envLabel newEnv
        label1 = emitLabel labelNo
    in
      EmitResult
      (EmitNone,
       condCode ++
       emitResultCode thenResult ++
       [(FlagResetDontCare, ""),
        (OperandString1 "JUMP" label1, "if else jump")] ++
       emitLabels condLabels "else" ++ 
       emitResultCode elseResult ++
       [(Label label1, "if else end")],
       incEnvLabel newEnv)
emitStatementIfElse (EmitResultAlternative der []) thenStatements elseStatements = emitStatementIfElse der thenStatements elseStatements
emitStatementIfElse (EmitResultAlternative der (er:ers)) thenStatements elseStatements = emitStatementIfElse (EmitResultAlternative der ers) thenStatements elseStatements

emitStatementWhile (EmitResult (condVal, condCode, condEnv)) statements =
    let
        contentResult = emitStatement statements condEnv
        newEnv = emitResultEnvironment contentResult
        labelNo = envLabel newEnv
        label1 = emitLabel labelNo
        label2 = emitLabel (labelNo + 1)
    in
      EmitResult
      (EmitNone,
       [(Label label1, "while start")] ++
       condCode ++ toRegisterA condVal ++
       [(OperandInt1 "AIA" 0xf, "while judege"),
        (OperandInt1 "AIA" 1, ""),
        (OperandString1 "JUMP" label2, "")] ++
       emitResultCode contentResult ++
       [(FlagResetDontCare, ""),
        (OperandString1 "JUMP" label1, "while loop jump"),
        (Label label2, "while end")],
       (incEnvLabel . incEnvLabel) newEnv)
emitStatementWhile (EmitResultAlternative der (EmitResult (EmitValueFalseLabels condLabels, condCode, condEnv):ers)) statements =
    let
        contentResult = emitStatement statements condEnv
        newEnv = emitResultEnvironment contentResult
        labelNo = envLabel newEnv
        label1 = emitLabel labelNo
    in
      EmitResult
      (EmitNone,
       [(Label label1, "while start")] ++
       condCode ++ 
       emitResultCode contentResult ++
       [(FlagResetDontCare, ""),
        (OperandString1 "JUMP" label1, "while loop jump")] ++
       emitLabels condLabels "while end",
       incEnvLabel newEnv)
emitStatementWhile (EmitResultAlternative der []) statements = emitStatementWhile der statements
emitStatementWhile (EmitResultAlternative der (er:ers)) statements = emitStatementWhile (EmitResultAlternative der ers) statements

emitExpression tree env = 
    case tree of
      (Number i) -> EmitResult (EmitIntValueConstant i, [], env)
      (VariableValue (Variable s)) -> EmitResult (EmitIntVariable (offsetTable (envTable env) s) s, [], env)

      PreOperation1 ("+" , expression) -> emitExpression expression env

      PreOperation1 ("-" , (Number i)) -> EmitResult (EmitIntValueConstant (mod (0xf + 1 - i) 0x10), [], env)

      PreOperation1 (operator , expression)
          | any (==operator) ["++", "--"] ->
              let
                  emitResult = emitExpression expression env
                  newEnv = emitResultEnvironment emitResult
              in
                EmitResult(EmitIntValueRegisterA,
                           emitResultCode emitResult ++
                           addressToRegisterY (emitResultValue emitResult) ++
                           [(Operand0 "MA", ""),
                            (OperandInt1 "AIA" (if operator == "++" then 1 else 0xf), "Unary operator '" ++ operator ++ "'"),
                            (Operand0 "AM", "")],
                           newEnv)

      PreOperation1 ("&" , expression) ->
          let
              emitResult = emitExpression expression env
              newEnv = emitResultEnvironment emitResult
          in
            EmitResult(EmitAddressRegisterY,
                       emitResultCode emitResult ++
                       (addressToRegisterY (emitResultValue emitResult)),
                       newEnv)
                                          
      PreOperation1 ("!" , expression) -> emitExpressionNot (emitExpression expression env) 

      PreOperation1 (operator , expression) ->
          let
              emitResult = emitExpression expression env
              newEnv = emitResultEnvironment emitResult
          in
            case operator of
              "-" ->
                  EmitResult(EmitIntValueRegisterA,
                             emitResultCode emitResult ++
                             (toRegisterA (emitResultValue emitResult)) ++
                             [(FlagResetRegistersSafe, ""),
                              (OperandString1 "CAL" "CMPL", "Unary operator '" ++ operator ++ "'"),
                              (OperandInt1 "AIA" 1, "")],
                             newEnv)
              _ -> emitError ("unary operator " ++ operator ++ " not supported")
                  
      PostOperation1 (operator , expression)
          | any (==operator) ["++", "--"] ->
              let
                  emitResult = emitExpression expression env
                  newEnv = emitResultEnvironment emitResult
              in
                EmitResult(EmitIntValueRegisterA,
                           emitResultCode emitResult ++
                           addressToRegisterY (emitResultValue emitResult) ++
                           [(Operand0 "MA", ""),
                            (OperandInt1 "AIA" (if operator == "++" then 1 else 0xf), "Unary operator '" ++ operator ++ "'"),
                            (Operand0 "AM", ""),
                            (OperandInt1 "AIA" (if operator == "++" then 0xf else 1), "cancel")],
                           newEnv)
      PostOperation1 (operator , expression) -> emitError ("unary operator " ++ operator ++ " not supported")

      Operation2 ("&&" , expression1, expression2) -> emitExpressionAnd expression1 expression2 env
      Operation2 ("||" , expression1, expression2) -> emitExpressionOr expression1 expression2 env
                                                
      Operation2 (operator , expression1, expression2) ->
          let
              emitResult1 = emitExpression expression1 env
              emitResult2 = emitExpression expression2 (incEnvOffset (emitResultEnvironment emitResult1))
              newEnv = decEnvOffset (emitResultEnvironment emitResult2)
              offset = envOffset newEnv

              reverseEmitResult1 = emitExpression expression2 env
              reverseEmitResult2 = emitExpression expression1 (incEnvOffset (emitResultEnvironment reverseEmitResult1))
              reverseNewEnv = decEnvOffset (emitResultEnvironment reverseEmitResult2)
              reverseOffset = envOffset reverseNewEnv
          in
            case operator of
              o | any (o==) ["+", "-", "!="] -> let op = if o == "!=" then "-" else o in
                  EmitResult(EmitIntValueRegisterA,
                             emitResultCode emitResult1 ++
                             (toRegisterA (emitResultValue emitResult1)) ++
                             [(OperandInt1 "TIY" offset, ""),
                              (Operand0 "AM", "")] ++
                             emitResultCode emitResult2 ++
                             (toRegisterA (emitResultValue emitResult2)) ++
                             [(OperandInt1 "TIY" offset, ""),
                              (Operand0 ("M" ++ op), "Operator '" ++ operator ++ "'")],
                             newEnv)

              o | any (o==) [">=", "<=", ">", "<"] ->
                  let
                      emitResultOp1 = (if o == ">=" || o == "<" then emitResult1 else reverseEmitResult1)
                      emitResultOp2 = (if o == ">=" || o == "<" then emitResult2 else reverseEmitResult2)
                      targetNewEnv = (if o == ">=" || o == "<" then newEnv else reverseNewEnv)
                                      
                      labelNo = envLabel targetNewEnv
                      label1 = emitLabel labelNo
                      label2 = emitLabel (labelNo + 1)
                  in
                    EmitResultAlternative
                    (
                     EmitResult
                     (
                      EmitIntValueRegisterA,
                      emitResultCode emitResultOp1 ++
                      (toRegisterA (emitResultValue emitResultOp1)) ++
                      [(OperandInt1 "TIY" offset, ""),
                       (Operand0 "AM", "")] ++
                      emitResultCode emitResultOp2 ++
                      (toRegisterA (emitResultValue emitResultOp2)) ++
                      [(OperandInt1 "TIY" offset, ""),
                       (Operand0 "CH", ""),
                       (if o == "<=" || o == ">="
                        then (OperandInt1 "TIA" 0, "set false")
                        else (OperandInt1 "TIA" 1, "set true")
                       ),
                       (Operand0 "CH", ""),
                       (Operand0 ("M-"), "Operator '" ++ operator ++ "'"),
                       (OperandString1 "JUMP" label1, ""),
                       (if o == "<=" || o == ">="
                        then (OperandInt1 "TIA" 1, "set true")
                        else (OperandInt1 "TIA" 0, "set false")
                       ),
                       (Operand0 "CH", ""),
                       (Label label1, ""),
                       (Operand0 "CH", "")],
                      incEnvLabel targetNewEnv
                     )
                    )

                    (emitResultCodeSizeSort
                     [
                      EmitResult
                      (
                       EmitValueFalseLabels [label1],
                       emitResultCode emitResultOp1 ++
                       (toRegisterA (emitResultValue emitResultOp1)) ++
                       [(OperandInt1 "TIY" offset, ""),
                        (Operand0 "AM", "")] ++
                       emitResultCode emitResultOp2 ++
                       (toRegisterA (emitResultValue emitResultOp2)) ++
                       [(OperandInt1 "TIY" offset, ""),
                        (Operand0 ("M-"), "Operator '" ++ operator ++ "'")] ++
                       (if o == "<" || o == ">" then [(OperandString1 "JUMP" label2, "")]  else []) ++
                       [(OperandString1 "JUMP" label1, "OOCV")] ++
                       (if o == "<" || o == ">" then [(Label label2, "")]  else []),
                       (incEnvLabel.incEnvLabel) targetNewEnv
                      ),
                      EmitResult
                      (
                       EmitValueTrueLabels [label1],
                       emitResultCode emitResultOp1 ++
                       (toRegisterA (emitResultValue emitResultOp1)) ++
                       [(OperandInt1 "TIY" offset, ""),
                        (Operand0 "AM", "")] ++
                       emitResultCode emitResultOp2 ++
                       (toRegisterA (emitResultValue emitResultOp2)) ++
                       [(OperandInt1 "TIY" offset, ""),
                        (Operand0 ("M-"), "Operator '" ++ operator ++ "'")] ++
                       (if o == "<=" || o == ">=" then [(OperandString1 "JUMP" label2, "")]  else []) ++
                       [(OperandString1 "JUMP" label1, "OOCV")] ++
                       (if o == "<=" || o == ">=" then [(Label label2, "")]  else []),
                       (incEnvLabel.incEnvLabel) targetNewEnv
                      )
                     ])
              "==" ->
                  EmitResultAlternative
                  (
                   EmitResult
                   (
                    EmitIntValueRegisterA,
                    emitResultCode emitResult1 ++
                    (toRegisterA (emitResultValue emitResult1)) ++
                    [(OperandInt1 "TIY" offset, ""),
                     (Operand0 "AM", "")] ++
                    emitResultCode emitResult2 ++
                    (toRegisterA (emitResultValue emitResult2)) ++
                    [(OperandInt1 "TIY" offset, ""),
                     (FlagResetRegistersSafe, ""),
                     (OperandString1 "CAL" "CHNG", "set false Ar'"),
                     (OperandInt1 "TIA" 0, ""),
                     (OperandInt1 "TIY" offset, ""),
                     (FlagResetRegistersSafe, ""),
                     (OperandString1 "CAL" "CHNG", ""),
                     (Operand0 ("M-"), "Operator '" ++ operator ++ "'"),
                     (OperandInt1 "AIA" 0xf, ""),
                     (OperandString1 "CAL" "CHNG", "")],
                    newEnv
                   )
                  )
                  [
                   EmitResult
                   (
                    EmitValueFalseLabels [(emitLabel (envLabel newEnv))],
                    case (emitResultValue emitResult1, emitResultValue emitResult2) of
                      ((EmitIntValueConstant constant), _) ->
                          emitResultCode emitResult2 ++
                          (toRegisterA (emitResultValue emitResult2)) ++
                          [(OperandInt1 "CIA" constant, "Compare constant " ++ show constant ++ " OOCMP=="),
                           (OperandString1 "JUMP" (emitLabel (envLabel newEnv)), "")]
                      (_, (EmitIntValueConstant constant)) ->
                          emitResultCode emitResult1 ++
                          (toRegisterA (emitResultValue emitResult1)) ++
                          [(OperandInt1 "CIA" constant, "Compare constant " ++ show constant ++ " OOCMP=="),
                           (OperandString1 "JUMP" (emitLabel (envLabel newEnv)), "")]
                      _ ->
                          emitResultCode emitResult1 ++
                          (toRegisterA (emitResultValue emitResult1)) ++
                          [(OperandInt1 "TIY" offset, ""),
                           (Operand0 "AM", "")] ++
                          emitResultCode emitResult2 ++
                          (toRegisterA (emitResultValue emitResult2)) ++
                          [(OperandInt1 "TIY" offset, ""),
                           (Operand0 ("M-"), "Operator '" ++ operator ++ "'"),
                           (OperandInt1 "CIA" 0, ""),
                           (OperandString1 "JUMP" (emitLabel (envLabel newEnv)), "OOCV")]
                   ,
                    incEnvLabel newEnv
                   )
                  ]

              _ -> emitError ("operator " ++ operator ++ " not supported")

      (ExpressionCall0 fname) -> emitExpressionCall0 fname env
      (ExpressionCall1 (fname, expression)) -> emitExpressionCall1 fname expression env

      _ -> emitError ("unkown expression")

emitExpressionNot (EmitResult (expVal, expCode, expEnv)) =
    let
        labelNo = envLabel expEnv
        label1 = emitLabel labelNo
        label2 = emitLabel (labelNo + 1)
    in        
      (EmitResultAlternative
       (EmitResult
        (EmitIntValueRegisterA,
         expCode ++ toRegisterA expVal ++
         [(OperandInt1 "CIA" 0, "Operator !"),
          (OperandString1 "JUMP" label1, ""),
          (OperandInt1 "TIA" 1, ""),
          (FlagResetRegistersSafe, ""),
          (OperandString1 "JUMP" label2, ""),
          (Label label1, ""),
          (OperandInt1 "TIA" 0, ""),
          (Label label2, "")],
         (incEnvLabel . incEnvLabel) expEnv)
       )
       [
        EmitResult
        (EmitValueFalseLabels [label1],
         expCode ++ toRegisterA expVal ++
         [(OperandInt1 "CIA" 0, "Operator !"),
          (OperandString1 "JUMP" label1, "OOCV")],
         incEnvLabel expEnv),
        
        EmitResult
        (EmitValueTrueLabels [label1],
         expCode ++ toRegisterA expVal ++
         [(OperandInt1 "AIA" 0xf, "Operator !"),
          (OperandInt1 "AIA" 1, ""),
          (OperandString1 "JUMP" label1, "OOCV")],
         incEnvLabel expEnv)
       ]
      )

emitExpressionNot (EmitResultAlternative der ers) =
    (EmitResultAlternative (emitExpressionNot der) (map reverseEmitValueLabel ers))

reverseEmitValueLabel (EmitResult (EmitValueFalseLabels expLabels, expCode, expEnv)) = (EmitResult (EmitValueTrueLabels expLabels, expCode, expEnv))
reverseEmitValueLabel (EmitResult (EmitValueTrueLabels expLabels, expCode, expEnv)) = (EmitResult (EmitValueFalseLabels expLabels, expCode, expEnv))

emitExpressionAnd expression1 expression2 env =
    let
        emitResult1 = emitResultSelect emitResultIsFalseLabels (emitExpression expression1 env)
        emitResult2 = emitResultSelect emitResultIsFalseLabels (emitExpression expression2 (emitResultEnvironment emitResult1))
        expVal1 = emitResultValue emitResult1
        expVal2 = emitResultValue emitResult2
        expCode1 = emitResultCode emitResult1
        expCode2 = emitResultCode emitResult2
        newEnv = emitResultEnvironment emitResult2
        labelNo = envLabel newEnv
        label1 = emitLabel labelNo
    in
      case (expVal1, expVal2) of
        (EmitValueFalseLabels expLabels1, EmitValueFalseLabels expLabels2) ->
            (EmitResultAlternative
             (EmitResult
              (EmitIntValueRegisterA,
               expCode1 ++ 
               expCode2 ++ 
               [(OperandInt1 "TIA" 1, "Operator &&"),
                (FlagResetRegistersSafe, ""),
                (OperandString1 "JUMP" label1, "")] ++
               emitLabels (expLabels1 ++ expLabels2) "" ++
               [(OperandInt1 "TIA" 0, ""),
                (Label label1, "")],
               incEnvLabel newEnv)
             )
             [
              EmitResult
              (EmitValueFalseLabels (expLabels1 ++ expLabels2),
               expCode1 ++
               expCode2,
               newEnv)
             ]
            )
            
        _ ->
            let
                emitResult1 = emitExpression expression1 env
                emitResult2 = emitExpression expression2 (emitResultEnvironment emitResult1)
                expVal1 = emitResultValue emitResult1
                expVal2 = emitResultValue emitResult2
                expCode1 = emitResultCode emitResult1
                expCode2 = emitResultCode emitResult2
                newEnv = emitResultEnvironment emitResult2
                labelNo = envLabel newEnv
                label1 = emitLabel labelNo
                label2 = emitLabel (labelNo + 1)
            in
              (EmitResultAlternative
               (EmitResult
                (EmitIntValueRegisterA,
                 expCode1 ++ toRegisterA expVal1 ++
                 [(OperandInt1 "AIA" 0xf, "Operator && 1"),
                  (OperandInt1 "AIA" 1, ""),
                  (OperandString1 "JUMP" label1, "")] ++
                 expCode2 ++ toRegisterA expVal2 ++
                 [(OperandInt1 "AIA" 0xf, "Operator && 2"),
                  (OperandInt1 "AIA" 1, ""),
                  (OperandString1 "JUMP" label1, ""),
                  (OperandInt1 "TIA" 1, ""),
                  (Label label1, "")],
                 incEnvLabel newEnv)
               )
               [
                EmitResult
                (EmitValueFalseLabels [label1],
                 expCode1 ++ toRegisterA expVal1 ++
                 [(OperandInt1 "AIA" 0xf, "Operator && 1"),
                  (OperandInt1 "AIA" 1, ""),
                  (OperandString1 "JUMP" label1, "OOCV")] ++
                 expCode2 ++ toRegisterA expVal2 ++
                 [(OperandInt1 "AIA" 0xf, "Operator && 2"),
                  (OperandInt1 "AIA" 1, ""),
                  (OperandString1 "JUMP" label1, "")],
                 incEnvLabel newEnv),
                
                EmitResult
                (EmitValueTrueLabels [label2],
                 expCode1 ++ toRegisterA expVal1 ++
                 [(OperandInt1 "AIA" 0xf, "Operator && 1"),
                  (OperandInt1 "AIA" 1, ""),
                  (OperandString1 "JUMP" label1, "OOCV")] ++
                 expCode2 ++ toRegisterA expVal2 ++
                 [(OperandInt1 "AIA" 0xf, "Operator && 2"),
                  (OperandInt1 "AIA" 1, ""),
                  (OperandString1 "JUMP" label1, ""),
                  (OperandString1 "JUMP" label2, ""),
                  (Label label1, "")],
                 (incEnvLabel.incEnvLabel) newEnv)
               ]
              )

emitExpressionOr expression1 expression2 env =
    let
        emitResult1 = emitResultSelect emitResultIsTrueLabels (emitExpression expression1 env)
        emitResult2 = emitResultSelect emitResultIsTrueLabels (emitExpression expression2 (emitResultEnvironment emitResult1))
        expVal1 = emitResultValue emitResult1
        expVal2 = emitResultValue emitResult2
        expCode1 = emitResultCode emitResult1
        expCode2 = emitResultCode emitResult2
        newEnv = emitResultEnvironment emitResult2
        labelNo = envLabel newEnv
        label1 = emitLabel labelNo
        label2 = emitLabel (labelNo + 1)
    in
      case (expVal1, expVal2) of

        (EmitValueTrueLabels expLabels1, EmitValueTrueLabels expLabels2) ->
            (EmitResultAlternative
             (EmitResult
              (EmitIntValueRegisterA,
               expCode1 ++ 
               expCode2 ++ 
               [(OperandInt1 "TIA" 0, "Operator || D"),
                (FlagResetRegistersSafe, ""),
                (OperandString1 "JUMP" label1, "")] ++
               emitLabels (expLabels1 ++ expLabels2) "" ++
               [(OperandInt1 "TIA" 1, ""),
                (Label label1, "")],
               incEnvLabel newEnv)
             )
             [
              EmitResult
              (EmitValueTrueLabels (expLabels1 ++ expLabels2),
               expCode1 ++
               expCode2,
               newEnv),

              EmitResult
              (EmitValueFalseLabels [label1],
               expCode1 ++
               expCode2 ++
               [(FlagResetRegistersSafe, ""),
                (OperandString1 "JUMP" label1, "Operator || E")] ++
               emitLabels (expLabels1 ++ expLabels2) "",
               incEnvLabel newEnv)
             ]
            )

        _ ->
            let
                emitResult1 = emitExpression expression1 env
                emitResult2 = emitExpression expression2 (emitResultEnvironment emitResult1)
                expVal1 = emitResultValue emitResult1
                expVal2 = emitResultValue emitResult2
                expCode1 = emitResultCode emitResult1
                expCode2 = emitResultCode emitResult2
                newEnv = emitResultEnvironment emitResult2
                labelNo = envLabel newEnv
                label1 = emitLabel labelNo
                label2 = emitLabel (labelNo + 1)
            in
              (EmitResultAlternative
               (EmitResult
                (EmitIntValueRegisterA,
                 expCode1 ++ toRegisterA expVal1 ++
                 [(OperandInt1 "CIA" 0, "Operator || 1 A"),
                  (OperandString1 "JUMP" label1, "")] ++
                 expCode2 ++ toRegisterA expVal2 ++
                 [(OperandInt1 "CIA" 0, "Operator || 2 A"),
                  (OperandString1 "JUMP" label1, ""),
                  (OperandInt1 "TIA" 0, ""),
                  (Label label1, "")],
                 incEnvLabel newEnv)
               )
               [
                EmitResult
                (EmitValueTrueLabels [label1],
                 expCode1 ++ toRegisterA expVal1 ++
                 [(OperandInt1 "CIA" 0, "Operator || 1 C"),
                  (OperandString1 "JUMP" label1, "OOCV")] ++
                 expCode2 ++ toRegisterA expVal2 ++
                 [(OperandInt1 "CIA" 0, "Operator || 2 C"),
                  (OperandString1 "JUMP" label1, "")],
                 incEnvLabel newEnv),

                EmitResult
                (EmitValueFalseLabels [label2],
                 expCode1 ++ toRegisterA expVal1 ++
                 [(OperandInt1 "CIA" 0, "Operator || 1 B"),
                  (OperandString1 "JUMP" label1, "OOCV")] ++
                 expCode2 ++ toRegisterA expVal2 ++
                 [(OperandInt1 "CIA" 0, "Operator || 2 B"),
                  (OperandString1 "JUMP" label1, ""),
                  (FlagResetDontCare, ""),
                  (OperandString1 "JUMP" label2, ""),
                  (Label label1, "")],
                 (incEnvLabel.incEnvLabel) newEnv)
               ]
              )

      
emitDeclarations declarations env = foldl (\x y -> emitDeclaration y x) env declarations

emitDeclaration (Declaration (dtype, variables)) env =
    foldl (\x y -> emitDeclarationVariable dtype y x) env variables

emitDeclarationVariable TypeInt (Variable vname) env =
    if offset < 0 then
        emitError "memory over"
    else
        Environment (newTable, offset - 1, envLabel env)
    where
      offset = envOffset env
      newTable = updateTable (envTable env) vname offset

emitExpressionCall0 fname env = 
    case fname of
      "exit" ->
          let
              labelNo = envLabel env
              label1 = emitLabel labelNo
          in
            EmitResult(EmitVoidValue,
                       [(FlagResetDontCare, ""),
                        (Label label1, ""),
                        (OperandString1 "JUMP" label1, "exit loop")],
                       incEnvLabel env)
          
      _ ->
          EmitResult (EmitVoidValue, [(FlagResetDontCare, "")] ++ emitFunctionCall0 fname, env)

emitFunctionCall0 fn@"resetNumberLed" = [(OperandString1 "CAL" "RSTO", "call " ++ fn ++ "()")]
emitFunctionCall0 fn@"beepEndSound" = [(OperandString1 "CAL" "ENDS", "call " ++ fn ++ "()")]
emitFunctionCall0 fn@"beepErrorSound" = [(OperandString1 "CAL" "ERRS", "call " ++ fn ++ "()")]
emitFunctionCall0 fn@"beepShortSound" = [(OperandString1 "CAL" "SHTS", "call " ++ fn ++ "()")]
emitFunctionCall0 fn@"beepLongSound" = [(OperandString1 "CAL" "LONS", "call " ++ fn ++ "()")]
emitFunctionCall0 fname = emitError ("unkown function " ++ fname ++ "(void)")

emitExpressionCall1 fname expression env = 
    case fname of
      "inputKey" ->
          let
              emitResult = emitExpression expression env
              newEnv = emitResultEnvironment emitResult
              addressCalcCode = emitResultCode emitResult
              labelNo = envLabel newEnv
              label1 = emitLabel labelNo
              label2 = emitLabel (labelNo + 1)
          in
            case addressCalcCode of
              [(OperandInt1 "TIY" _, _)] ->
                  EmitResultAlternative
                  (
                   EmitResult
                   (EmitIntValueRegisterA,
                    [(OperandInt1 "TIA" 0, "inputKey start"),
                     (Operand0 "CH" , ""),
                     (Operand0 "KA", ""),
                     (OperandString1 "JUMP" label1, "")] ++
                    addressCalcCode ++
                    [(Operand0 "AM", ""),
                     (OperandInt1 "TIA" 1, ""),
                     (Operand0 "CH" , ""),
                     (Label label1, ""),
                     (Operand0 "CH" , "")],
                    incEnvLabel newEnv
                   )
                  )
                  [
                   EmitResult
                   (
                    EmitValueFalseLabels [label1],
                    [(Operand0 "KA", ""),
                     (OperandString1 "JUMP" label1, "OOCV")] ++
                    addressCalcCode ++
                    [(Operand0 "AM", "")],
                    incEnvLabel newEnv
                   ),
                   EmitResult
                   (
                    EmitValueTrueLabels [label2],
                    [(Operand0 "KA", ""),
                     (OperandString1 "JUMP" label1, "OOCV")] ++
                    addressCalcCode ++
                    [(Operand0 "AM", ""),
                     (FlagResetDontCare, ""),
                     (OperandString1 "JUMP" label2, ""),
                     (Label label1, "")],
                    (incEnvLabel.incEnvLabel) newEnv
                   )
                  ]

              _ -> emitError ("can't emit expression in inputKey(..), code size too large")
      _ ->
          let
              emitResult = emitExpression expression env
          in
            EmitResult(EmitVoidValue,
                       emitResultCode emitResult ++
                       toRegisterA (emitResultValue emitResult) ++
                       emitFunctionCall1 fname,
                       emitResultEnvironment emitResult)

emitFunctionCall1 fn@"setNumberLed" = [(Operand0 "AO", "call " ++ fn ++ "()")]
emitFunctionCall1 fn@"setLed" = [(Operand0 "CY", ""), (OperandString1 "CAL" "SETR", "call " ++ fn ++ "()")]
emitFunctionCall1 fn@"resetLed" = [(Operand0 "CY", ""), (OperandString1 "CAL" "RSTR", "call " ++ fn ++ "()")]
emitFunctionCall1 fn@"beep" = [(FlagResetRegistersSafe, "")] ++ [(OperandString1 "CAL" "SUND", "call " ++ fn ++ "()")]
emitFunctionCall1 fn@"sleep" = [(FlagResetRegistersSafe, "")] ++ [(OperandString1 "CAL" "TIMR", "call " ++ fn ++ "()")]
emitFunctionCall1 fname = emitError ("unkown function " ++ fname ++ "(int)")

toRegisterA EmitIntValueRegisterA = []
toRegisterA (EmitIntValueTemporary i) =
    [(OperandInt1 "TIY" i, "Temporary offset " ++ show i) , (Operand0 "MA", "")]
toRegisterA (EmitIntVariable i s) =
    [(OperandInt1 "TIY" i, "Variable '" ++ s ++ "' (offset " ++ show i ++ ")") , (Operand0 "MA", "")]
toRegisterA (EmitIntValueConstant i) = 
    [(OperandInt1 "TIA" i, "Constant " ++ show i)]
toRegisterA EmitAddressRegisterY = 
    [(Operand0 "CY", "")]
toRegisterA EmitVoidValue = emitError ("void value not ignored as it ought to be")

addressToRegisterY EmitAddressRegisterY = []
addressToRegisterY (EmitIntVariable i s) =
    [(OperandInt1 "TIY" i, "Variable '" ++ s ++ "' (offset " ++ show i ++ ")")]
addressToRegisterY t = emitError ("invalid reference " ++ show t)

emitLabel = ("LABEL" ++) . show

emitLabels (l:ls) comment = (Label l, comment) : emitLabels (ls) ""
emitLabels [] _ = []

emitResultCodeSizeSort = sortBy emitResultCodeSizeCompare
emitResultCodeSizeCompare (EmitResult (_, a, _)) (EmitResult (_, b, _)) =
    compare (sum (map (instructionSize.fst) a)) (sum (map (instructionSize.fst) b))
